## file: Rcodes-chap04-eval.r
## usage: source('Rcodes-chap04-eval.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(Bolstad)                   # sintegral
require(doParallel)                # foreach(並列計算)
cl <- makeCluster(detectCores())   # クラスタの作成
registerDoParallel(cl)


## テスト誤差とトレーニング誤差のプロット
par <- seq(-3,3,l=50) # パラメータの範囲
te_err <- (1+par^2)/2 # テスト誤差
## テスト誤差をプロット
plot(par,te_err, lwd=4, lty=1,type='l')
for(i in 1:10){
  z <- rnorm(20)                                 # データ生成
  tr_err <- colMeans(outer(z,par,FUN="-")^2/2)   # トレーニング誤差
  lines(par,tr_err,lty=2,type='l')               # トレーニング誤差をプロット
}


readline(msg)
## スプラインに対する交差検証法
n <- 100; K <- 10                  # 設定：データ数．10重CV
## データ生成 
x <- runif(n,min=-2,max=2)         # 区間[-2,2]上の一様分布
y <- sin(2*pi*x)/x + rnorm(n,sd=0.5)
cv_idx  <- rep(1:K,ceiling(n/K))[1:n]        # データをグループ分け 
lambdas <- 10^(-10:1)                        # λの候補
cv_err <- foreach(l=lambdas,.combine=c,.packages="foreach")%dopar%{# 並列計算
  ## 各lambdaに対してK回反復
  err <- foreach(k=1:K,.combine=c)%dopar%{
    tr_idx <- which(cv_idx!=k)               # D^(k)
    te_idx <- which(cv_idx==k)               # D_k
    cvx <- x[tr_idx]
    cvy <- y[tr_idx]
    ## 検証用データで学習
    res <- smooth.spline(cvx, cvy, lambda=l)
    ## テスト誤差の推定 
    mean((y[te_idx]-predict(res,x[te_idx])$y)^2/2)
  }
  mean(err)
}
## λに対する検証誤差のプロット
plot(lambdas,cv_err,log='x') 
stopCluster(cl)


readline(msg)
## データからAUCを計算(n=100)
n <- 100                                       # データ数 100
xp <- matrix(rnorm(n*2,mean=1),n)              # 信号アリ
xn <- matrix(rnorm(n*2),n)                     # 信号ナシ
mean(outer(xp[,1],xn[,1],FUN=">="))            # F1 のAUC
mean(outer(rowSums(xp),rowSums(xn),FUN=">="))  # F2 のAUC


readline(msg)
## データからAUCを計算(n=1000)
n <- 10000                                     # データ数 10000
xp <- matrix(rnorm(n*2,mean=1),n)              # 信号アリ
xn <- matrix(rnorm(n*2),n)                     # 信号ナシ
mean(outer(xp[,1],xn[,1],FUN=">="))            # F1 のAUC
mean(outer(rowSums(xp),rowSums(xn),FUN=">="))  # F2 のAUC


readline(msg)
## 数値積分でAUCを計算

c <- seq(-10,10,by=0.01)
## F1
fpr <- function(c){1-pnorm(c)}
tpr <- function(c){1-pnorm(c,mean=1)}
sintegral(fpr(c),tpr(c))$value        # 数値積分によるAUCの計算
## F2
fpr <- function(c){1-pnorm(c,sd=sqrt(2))}
tpr <- function(c){1-pnorm(c,mean=2,sd=sqrt(2))}
sintegral(fpr(c),tpr(c))$value        # 数値積分によるAUCの計算
