## file: Rcodes-chap05-dataDescription.r 
## usage: source('Rcodes-chap05-dataDescription.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(car)         # Davis データ
require(mlbench)     # BostonHousing データ
require(MASS)        # isoMDS
require(HSAUR3)      # voting データ


## 主成分分析
data(Davis)                    # データ読み込み
plot(Davis[,c(2,3)])           # データのプロット
## i=12のデータは外れ値として除去．
## 身長の単位を[メートル]に変換し，対数の値を計算．
d <- log(cbind(Davis[-12,2],Davis[-12,3]/100))
pca <- prcomp(d); pca          # 主成分分析を実行，結果の表示
summary(pca)                   # 結果のまとめを表示


readline(msg)
## 因子分析
data(BostonHousing)  # データ読み込み
# ?BostonHousing       # 各変数の説明


readline(msg)
## カテゴリデータを数値化
BostonHousing$chas <- as.numeric(BostonHousing$chas)
## 因子数3，因子スコアをThomson法で計算
f <- factanal(BostonHousing[,-14],factors=3,scores='regression')
f$loadings    # 結果の表示
## 因子スコアのプロット
plot(data.frame(f$scores))


readline(msg)
## 多次元尺度構成法(MDS)
n <- 10                     # データ数
k <- 2                      # データの次元
V <- matrix(runif(n*k),n,k) # データ点の生成
d <- dist(V)                # 距離行列の計算
rV2 <- cmdscale(d,k=2)      # MDS(2次元点配置を再構成)
rV1 <- cmdscale(d,k=1)      # MDS(1次元点配置を再構成)
## 2次元点配置を再構成
par(mfrow=c(1,2))
plot( V, type="p", cex=2.4, main="点配置")           # 元のデータ点
text(V[,1],V[,2],labels=0:(n-1))
plot(rV2, type="p", cex=2.4, main="再構成した点配置") # MDSで再構成したデータ点
text(rV2[,1],rV2[,2],labels=0:(n-1))


readline(msg)
cmdscale(d,k=2, eig=T)$eig[1:3]           # MDS：行列 B の固有値
## MDS：マンハッタン距離(L1距離)
d <- dist(V, method="manhattan")
cmdscale(d,k=2,eig=T)$eig                 # MDS(マンハッタン距離)：行列 B の固有値


readline(msg)
## 非計量的MDS
data(voting)             # データ読み込み
nmMDS <- isoMDS(voting)  # 非計量的MDS(デフォルトでk=2)
nmMDS


readline(msg)
col <- c()                                     # 各点の色を設定
col[grep("(R)",rownames(nmMDS$points))] <- 2   # Republican Party
col[grep("(D)",rownames(nmMDS$points))] <- 4   # Democratic Party
## 色分けして数字を表示
plot(nmMDS$points,type='n',main="非計量的MDS")
text(nmMDS$points,col=col)


readline(msg)
## votingに計量的MDSを適用
mMDS <- cmdscale(voting,eig=T)                # 計量的MDS
col <- c()                                    # 各点の色を設定
col[grep("(R)",rownames(mMDS$points))] <- 2   # Republican Party
col[grep("(D)",rownames(mMDS$points))] <- 4   # Democratic Party
## 色分けして数字を表示
plot(mMDS$points,type='n',main="計量的MDS")
text(mMDS$points,col=col)
