## file: Rcodes-chap09-clustering.r
## usage: source('Rcodes-chap09-clustering.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(HDclassif)   # wine データ
require(kernlab)     # kkmeans
require(mlbench)     # mlbench.circle
require(HSAUR3)      # voting データ
require(mclust)      # Mclust


## k平均法: wineデータ
data(wine)          # データ読み込み
x <- wine[,-1]      # 品種の情報を除く．
x <- scale(x)       # 各軸をスケーリング
class <- wine[,1]   # 品種
km <- kmeans(x,3,nstart=10)   # k=3でk平均法を実行
km$cluster                    # 各データ点に対応するクラスタを表示
## 2次元にプロット
pca <- prcomp(x)          # 主成分分析       
par(mfrow=c(1,2))  
plot(pca$x[,1:2],col=class,main="wine: 品種によるグループ分け")    # 品種によるグループ分け
plot(pca$x[,1:2],col=km$cl,main="k平均法によるグループ分け") # クラスタリングによるグループ分け


readline(msg)
## k平均法: irisデータ
x <- scale(iris[,-5]) 
class <- iris[,5]
km <- kmeans(x,3,nstart=10)  # k-平均法 (k=3) 
## 2次元にプロット
pca <- prcomp(x)             # 主成分分析
par(mfrow=c(1,2))
plot(pca$x[,1:2],col=class,main="iris: 品種によるグループ分け")    # 種類によるグループ分け
plot(pca$x[,1:2],col=km$cl,main="k平均法によるグループ分け") # クラスタリングによるグループ分け


readline(msg)
## カーネルk平均法: mlbench.circleのクラスタリング
## データを生成
data <- mlbench.circle(200, d=2)  
data$x[data$cl==1,] <- 0.5*data$x[data$cl==1,]
## クラスタ数2のカーネルk平均法．ガウスカーネルを使う．
kkm <- kkmeans(data$x, 2, kernel="rbfdot") 
par(mfrow=c(1,1))
plot(data$x, col=kkm, main="カーネルk平均法") # プロット


readline(msg)
## スペクトラル・クラスタリング
## データの生成
data <- mlbench.spirals(300, cycles=1, sd=0.05)
## スペクトラル・クラスタリング
sc <- specc(data$x, centers=2)
par(mfrow=c(1,2))
plot(data$x, col=sc,main="スペクトラル・クラスタリング")     # プロット
##カーネルk-平均法
kkm <- kkmeans(data$x,2)
plot(data$x, col=kkm,main="カーネルk-平均法")


readline(msg)
## 局所性保存射影と多次元尺度構成法
data(voting)                        # データ読み込み
W <- exp(-voting/median(voting))    # 重みに変換(15次行列)
L <- diag(rowSums(W))-W             # グラフ・ラプラシアン
ev <- eigen(L)$vector[,13:14]       # 小さい固有値に対応する固有ベクトル
col <- c()                          # 各点の色を設定
col[grep("(R)",rownames(W))] <- 2   # Republican Party
col[grep("(D)",rownames(W))] <- 4   # Democratic Party
par(mfrow=c(1,1))
plot(ev,type='n',main="votingデータに局所性保存射影を適用")    # プロット(色分けして数字を表示)
text(ev,col=col)                    


readline(msg)
## 混合正規分布によるクラスタリング
x <- scale(wine[,-1])   # wine データ
class <- wine[,1]
fit <- Mclust(x, G=3)   # グループ数3でクラスタリング
pca <- prcomp(x)        # PCAで低次元に射影
par(mfrow=c(1,2))
plot(pca$x[,1:2], col=fit$class, main="グループ数=3") # プロット
fit <- Mclust(x)        # グループ数の推定してクラスタリング
fit$G                   # グループ数の推定結果
plot(pca$x[,1:2],col=fit$class, main=paste("グループ数の推定結果: ",fit$G))  # プロット


readline(msg)
data <- mlbench.spirals(300, cycles=1, sd=0.05)  # データ生成
fit <- Mclust(data$x)                            # グループ数の推定してクラスタリング
fit$G                                            # グループ数の推定結果
par(mfrow=c(1,1))                                # プロット
plot(data$x, col=fit$class, main=paste("グループ数の推定結果: ",fit$G))
