## file: Rcodes-chap11-sparse.r
## usage: source('Rcodes-chap11-sparse.r',echo=TRUE,max.deparse.length=Inf)
## required files: Rlogo.jpg, boat.png, goldhill.png, lena.png
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(glmnet)      # glmnet
require(genlasso)    # fusedlasso1d, fusedlasso2d
require(huge)        # huge, huge.generator
require(spams)       # spams
require(png)         # readPNG
require(jpeg)        # readJPEG


n <- 30       # データ数
d <- 50       # パラメータの次元
s <- 10       # パラメータの非ゼロ要素数
## データ生成
beta <- c(rep(1,s),rep(0,d-s))
X <- matrix(rnorm(n*d),n)
y <- X %*% beta + rnorm(n,sd=0.01)
##
la <- glmnet(X,y,alpha=1)                    # ラッソ
ri <- glmnet(X,y,alpha=0)                    # リッジ
## プロット
par(mfrow=c(1,2))
plot(la$b[1:d,ncol(la$b)],lwd=2,type='h',main="ラッソ")    # ラッソ
plot(ri$b[1:d,ncol(ri$b)],lwd=2,type='h',main="リッジ")    # リッジ


readline(msg)
## パス追跡のプロット
plot(la,main="ラッソ")       # ラッソ
plot(ri,main="リッジ")       # リッジ


readline(msg)
## エラスティックネット
n <- 100; d <- 200; s <- 30         # 設定：データ数, 次元, 非ゼロ要素数
beta <- c(rep(1,s),rep(0,d-s))      # 真のパラメータ
## データ生成
X <- matrix(rnorm(n*d),n)
y <- X %*% beta + rnorm(n,sd=0.01)
la <- glmnet(X,y,alpha=1)           # ラッソで学習
el <- glmnet(X,y,alpha=0.1)         # エラスティックネットで学習
## プロット
par(mfrow=c(1,1))
plot(colSums(abs(el$b)),colSums(el$b!=0),col=2,type='l',        # エラスティックネット
     lwd=2,xlab='L1 norm',ylab="非ゼロ要素の数")                     
lines(colSums(abs(la$b)),colSums(la$b!=0),lwd=2,lty=2)                 # ラッソ
lines(cbind(colSums(abs(el$b)),n),lty=3)
legend("topleft",c("エラスティックネット","ラッソ"),col=c(2,1),lty=c(1,2),lwd=c(2,2),bg="white")


readline(msg)
## フューズド・ラッソ
## データ生成
beta <- rnorm(5)%x%rep(1,10) 
y <- beta+rnorm(50,sd=0.2)
res <- fusedlasso1d(y)   # フューズド・ラッソによる推定
par(mfrow=c(1,2))
plot(y,main="データ")                            # データのプロット
lines(beta,col=2)
plot(res,main="フューズド・ラッソによる推定結果")   # フューズド・ラッソによる推定結果をプロット 


readline(msg)
## 2次元版フューズド・ラッソ
imgy <- readJPEG("Rlogo.jpg")       # 画像データ読み込み
z <- y <- t(apply(imgy[,,3],2,rev))                    # y: 元画像
i <- sample(prod(dim(y)),round(prod(dim(y))/5))        # ノイズの位置
z[i] <- pmin(pmax(z[i] + rnorm(length(i),sd=0.2),0),1) # z: 観測画像
image(y,col=gray((0:32)/32),main="元画像")              # 元画像のプロット
image(z,col=gray((0:32)/32),main="観測画像")            # 観測画像のプロット
readline(msg)
##    2次元版フューズド・ラッソの計算：時間がかかるため maxsteps=1000 で打ち切り
readline(msg)
system.time(res <- fusedlasso2d(z,maxsteps=1000,verbose=T)) # 2次元版フューズド・ラッソ
image(matrix(res$beta[,1000],nrow(y)),col=gray((0:32)/32),  # 学習結果の例をプロット
      main="maxsteps=1000 の結果") 
res$lambda[1000]                                            # このときの正則化パラメータの値
## res <- fusedlasso2d(z,maxsteps=Inf,verbose=T)  # λ=0 と等価になるまで計算(時間がかかる)
## length(res$lambda)                             # パス追跡で計算した正則化パラメータの個数
## image(matrix(res$beta[,12000],nrow(y)),col=gray((0:32)/32))   # 学習結果の例をプロット
## res$lambda[12000]                              # このときの正則化パラメータの値


readline(msg)
## スパース・ロジスティック回帰
n <- 100; d <- 200; s <- 30     ## 設定：データ数，次元，非ゼロ要素数
theta <- c(rep(1,s),rep(0,d-s)) ## 真のパラメータ設定
## データ生成
X <- matrix(rnorm(n*d),n)
y <- X%*%theta+rnorm(n,sd=0.001)>=0   
## パラメータ推定
la <- glmnet(X,y,family='binomial',alpha=1)    # L1 正則化
ri <- glmnet(X,y,family='binomial',alpha=0)    # L2 正則化
## プロット：いくつかのオプションは省略
par(mfrow=c(1,2))
plot(la$b[1:d,ncol(la$b)],type='h',main="ロジスティック回帰：L1 正則化")   # L1 正則化
plot(ri$b[1:d,ncol(ri$b)],type='h',main="ロジスティック回帰：L2 正則化")   # L2 正則化



readline(msg)
## グラフィカル・ラッソ
n <- 100; d <- 5                # 設定：データ数，次元
x <- huge.generator(n,d,p=0.5)  # 人工データ生成，pは辺を生成する確率
x$sparsity                      # 生成した精度行列のスパース度
x$theta                         # 精度行列のパターン


readline(msg)
## グラフィカル・ラッソ: 結果のプロット
n <- 1000; d <- 20                # 設定：データ数，次元
x <- huge.generator(n,d,p=0.05)   # 人工データ生成，pは辺を生成する確率
res <- huge(x$data)               # グラフィカル・ラッソによる推定
par(mfrow=c(1,1))
plot(res)                         # 結果をプロット


readline(msg)
## グラフィカル・ラッソを stockdata に適用
data(stockdata)       # データ読み込み
dim(stockdata$data)   # データのサイズ
res <- huge(stockdata$data)
plot(res)


readline(msg)
## 辞書学習
## 画像データ読み込み
A <- readPNG('boat.png')
B <- readPNG('goldhill.png')
C <- readPNG('lena.png')
## 3枚の画像から8x8の画像パッチを切り出す．それらをまとめてデータ行列 X を作成．
X <- cbind(spams.im2col_sliding(A,8,8),spams.im2col_sliding(B,8,8),spams.im2col_sliding(C,8,8))
dim(X)
## 画像パッチの表示
par(mfrow=c(10,10),mai=rep(0,4))
for(i in 1:100){ # ランダムに画像パッチを選んで表示
  image(matrix(X[,sample(ncol(X),1)],8),axes=FALSE,col=gray((0:32)/32))
}


readline(msg)
## 辞書学習
D <- spams.trainDL(X,K=100,lambda1=1)              # 辞書学習
for(i in 1:ncol(D)){                               # 辞書をプロット
  image(matrix(D[,i],8),axes=FALSE,col=gray((0:32)/32))
}
