## file: Rcodes-chap12-ensemble.r
## usage: source('Rcodes-chap12-ensemble.r',echo=TRUE,max.deparse.length=Inf)
msg <- "------------------------ Type  <Return> to continue ------------------------: "
require(rpart)             # rpart, stagecデータ
require(rpart.plot)        # rpart.plot
require(adabag)            # bagging
require(randomForest)      # randomForest
require(ada)               # ada
require(xgboost)           # xgboost
require(kernlab)           # spamデータ
require(doParallel)        # foreach (並列計算)
registerDoParallel(detectCores())   # クラスタの作成


stagec$pgstat <- as.factor(stagec$pgstat)   # データの出力ラベルを作成
## モデリング
fm <- as.formula(paste("pgstat~",paste(dimnames(stagec)[[2]][3:8],collapse='+')))
rp  <- rpart(fm,data=stagec)    # rpartで学習
rp$control$cp                   # αに対応する正則化パラメータのデフォルト値
## 決定木をプロット
par(mfrow=c(1,2))
plot(rp,main=paste("正則化パラメータ:",rp$control$cp)); text(rp,use.n=TRUE,all=TRUE,xpd=NA)
## 正則化パラメータを大きくしてrpartで学習
rq <- rpart(fm,data=stagec,control=rpart.control(cp=0.05))
## 決定木をプロット
plot(rq,main=paste("正則化パラメータ:",rq$control$cp)); text(rq,use.n=TRUE,all=TRUE,xpd=NA)



readline(msg)
## rpart の結果を rpart.plot でプロット
rpart.plot(rq,main="rpart.plot でプロット")       # プロット



readline(msg)
## バギング: iris データに適用
idx <- sample(nrow(iris),100)      # iris の100データでトレーニング，残りでテスト
ba <- bagging(Species~.,data=iris[idx,],mfinal=50)   # bagging: B=50
pred <- predict(ba,iris[-idx,])    # 予測ラベル
pred$error                         # テスト誤差
pred$confusion                     # 予測ラベルと真のラベルの対応



readline(msg)
## バギング: stagecデータに適用
## Bを変えてテスト誤差を比較
B <- c(10,50,100)                   # baggingでB=10,50,100を試す
stagec$pgstat <- as.factor(stagec$pgstat)    # stagecデータを学習
## モデリング
fm <- as.formula(paste("pgstat ~", paste(dimnames(stagec)[[2]][3:8],collapse='+')))
## 各学習法を20回並列計算．結果をtesterrに格納
system.time(                                              # 計算時間を計測
testerr <- foreach(i=1:20,.combine=rbind)%dopar%{
  idx <- sample(nrow(stagec),100)                         # データサンプリング
  rp  <- rpart(fm,data=stagec[idx,])                      # rpartで学習
  err <- mean(predict(rp,stagec[-idx,],type='c')!=stagec[-idx,'pgstat'])  # テスト誤差
  for(b in B){                                            # baggingで学習・テスト
    ba  <- bagging(fm,data=stagec[idx,],mfinal=b)
    err <- c(err,predict(ba,stagec[-idx,])$error)
  }
  err
})
## boxplotでプロット
par(mfrow=c(1,1))
dimnames(testerr)[[2]] <- c('rpart',paste('bagging',B,sep=''))
boxplot(testerr, main="rpart と bagging のテスト誤差")
colMeans(testerr)                                         # 結果を表示



readline(msg)
## ランダムフォレスト
stagec$pgstat <- as.factor(stagec$pgstat)  # データ生成
## モデリング
fm <- as.formula(paste("pgstat ~",paste(dimnames(stagec)[[2]][3:8],collapse='+')))
## ランダムフォレストで学習 (na.omitで NAデータを除く)
idx <- sample(nrow(stagec),100)
rf <- randomForest(fm, data=stagec[idx,], na.action=na.omit)
rf                                              # 結果を表示
predy <- predict(rf,stagec[-idx,])              # 検証データでラベル予測
mean(predy!=stagec[-idx,"pgstat"],na.rm=TRUE)   # 検証誤差
plot(rf,main="ランダムフォレストのテスト誤差，FNR(赤), FPR(緑)")      # 結果をプロット



readline(msg)
## ラベル予測：adaboost と XGBoost を比較
data(soldat)                       # データ読み込み
dim(soldat)     
## 5631サンプルで入力は72次元，出力ラベルは 0, 1 の2値
x <- soldat[,-73]; y <- soldat[,73]; y[y==-1] <- 0  
T <- 2000                          # ブースティングの反復回数は2000
mp <- 1                            # 弱学習器の木の深さは 1 (決定株)
idx <- sample(nrow(x),4000)
tr <- list(x=x[ idx,],y=y[ idx])    # トレーニングデータ
te <- list(x=x[-idx,],y=y[-idx])    # テストデータ
## アダブースト (計算時間を計測) 
## system.time(boE <- ada(tr$x,tr$y,loss='e',iter=T, control=rpart.control(maxdepth=mp))) # 時間がかかるのでコメントアウト
## mean(predict(boE,te$x)!=te$y)                                                          # アダブーストのテスト誤差
## xgboost (計算時間を計測)
system.time(boL <- xgboost(data=as.matrix(tr$x),label=tr$y,nround=T,
                           max.depth=mp,objective="binary:logistic",verbose=FALSE))
mean((predict(boL,as.matrix(te$x))>1/2)!=te$y) # xgboostのテスト誤差



readline(msg)
## ラベル予測：ランダムフォレストと XGBoost を比較
dat <- na.omit(soldat)              # NAを除く
x <- dat[,-73]; y <- dat[,73]; y[y==-1] <- 0  
idx <- sample(nrow(x),3000)
tr <- list(x=x[ idx,],y=y[ idx])    # トレーニングデータ
te <- list(x=x[-idx,],y=y[-idx])    # テストデータ
## ランダムフォレスト
system.time(rf <- randomForest(tr$x,as.factor(tr$y),ntree=1000))
## XGBoost
system.time(boL <- xgboost(data=as.matrix(tr$x),label=tr$y,nround=1000,objective="binary:logistic",verbose=0))
# テスト誤差
mean(predict(rf,te$x)!=te$y)        # ランダムフォレストのテスト誤差
mean((predict(boL,as.matrix(te$x))>1/2)!=te$y) # XGBoostのテスト誤差



readline(msg)
## 0-1損失に対する最適な反復数を交差検証で求める：soldat データ
T <- 5000                      # 反復数は最大5000  
mp <- 1                        # 弱学習器は決定株
K <- 5                         # 5-交差検証法
# 交差検証法 (計算時間を計測)
system.time(boLcv <- xgb.cv(data=as.matrix(tr$x),label=tr$y,nfold=K,nrounds=T,
                            max.depth=mp,metrics="error",objective="binary:logistic", verbose=F))
##
optT <- which.min(boLcv[[4]]$test_error_mean)               # 交差検証法から求めた最適な反復数で学習
boL  <- xgboost(data=as.matrix(tr$x),label=tr$y,nround=optT,
               max.depth=mp,objective="binary:logistic",verbose=FALSE)
testerr <- mean((predict(boL,as.matrix(te$x))>1/2)!=te$y); testerr              # テスト誤差
# トレーニング誤差と検証誤差をプロット
par(ps=16,mai=c(1,1,1,1))
ylim <- range(c(boLcv[[4]]$train_error_mean,boLcv[[4]]$test_error_mean))
 plot(boLcv[[4]]$train_error_mean,col=1,lty=1,lwd=2,xlab="反復数",ylab="誤差",ylim=ylim,type='l',main="soldat データ")
lines(boLcv[[4]]$test_error_mean, col=2,lty=2,lwd=2)
lines(cbind(0:T,testerr),lty=3,lwd=2)
legend('topright',c("検証誤差","トレーニング誤差","テスト誤差"),col=c(2,1,1),lwd=c(2,2,2),lty=c(2,1,3),y.intersp=1.4)



readline(msg)
## 0-1損失に対する最適な反復数を交差検証で求める：spam データ
data(spam)
x <- spam[,-58]; y <- as.numeric(spam[,58]=='spam')
idx <- sample(nrow(x),3000)
tr <- list(x=x[ idx,],y=y[ idx])  # トレーニングデータ
te <- list(x=x[-idx,],y=y[-idx])  # テストデータ
T <- 2000                         # 反復数は最大5000  
mp <- 5                           # 弱学習器は決定株
K <- 5                            # 5-交差検証法
## 交差検証法 (計算時間を計測)
system.time(boLcv <- xgb.cv(data=as.matrix(tr$x),label=tr$y,nfold=K,nrounds=T,
                            max.depth=mp,metrics="error",objective="binary:logistic", verbose=F))
##
optT <- which.min(boLcv[[4]]$test_error_mean)               # 交差検証法から求めた最適な反復数で学習
boL  <- xgboost(data=as.matrix(tr$x),label=tr$y,nround=optT,
               max.depth=mp,objective="binary:logistic",verbose=FALSE)
testerr <- mean((predict(boL,as.matrix(te$x))>1/2)!=te$y); testerr              # テスト誤差
## トレーニング誤差と検証誤差をプロット
par(ps=16,mai=c(1,1,1,1))
ylim <- range(c(boLcv[[4]]$train_error_mean,boLcv[[4]]$test_error_mean))
 plot(boLcv[[4]]$train_error_mean,col=1,lty=1,lwd=2,xlab="反復数",ylab="誤差",ylim=ylim,type='l',main="spam データ")
lines(boLcv[[4]]$test_error_mean, col=2,lty=2,lwd=2)
lines(cbind(0:T,testerr),lty=3,lwd=2)
legend('topright',c("検証誤差","トレーニング誤差","テスト誤差"),col=c(2,1,1),lwd=c(2,2,2),lty=c(2,1,3),y.intersp=1.4)



readline(msg)
## ブースティングにおける 0-1損失 と 対数損失の挙動
idx <- sample(nrow(x),4000)
tr <- xgb.DMatrix(data=as.matrix(x[ idx,]),label=y[ idx])    # トレーニングデータ
te <- xgb.DMatrix(data=as.matrix(x[-idx,]),label=y[-idx])    # テストデータ
watchlist <- list(train=tr, test=te)                         # 各反復での損失を計算
T <- 2000                                                    # 反復数
mp <- 5                                                      # 弱学習機は深さ 5 の決定木
boL <- xgb.train(data=tr,nround=T,max.depth=mp,watchlist=watchlist,objective="binary:logistic",
                 eval.metric="logloss",eval.metric="error",verbose=1,print_every_n=T)
## プロット
par(ps=16,mai=rep(1,4),mfrow=c(1,2))                                     
ylim <- range(c(boL$e[['train_error']],boL$e[['test_error']]))
 plot(boL$e[['train_error']],lwd=2,type='l',ylim=ylim,xlab='反復数',ylab='0-1損失')
lines(boL$e[['test_error']] ,lwd=2,col=2,lty=2)
legend('bottomright',c('トレーニングデータ','テストデータ'),col=c(1,2),lwd=c(2,2),lty=c(1,2),y.intersp=1.5,bg="white")
##
ylim <- range(c(boL$e[['train_logloss']],boL$e[['test_logloss']]))
 plot(boL$e[['train_logloss']],lwd=2,type='l',ylim=ylim,xlab='反復数',ylab='対数損失')
lines(boL$e[['test_logloss']] ,lwd=2,col=2,lty=2)
legend('bottomright',c('トレーニングデータ','テストデータ'),col=c(1,2),lwd=c(2,2),lty=c(1,2),y.intersp=1.5,bg="white")



readline(msg)
## 確率推定のための適切な反復数を交差検証法で求める．
T <- 2000                           # 反復数
mp <- 5                             # 弱学習機は深さ 5 の決定木
K <- 5                              # 5-重交差検証法
system.time(                        # 計算時間を計測
            boLcv <- xgb.cv(data=tr,nrounds=T,max.depth=mp,nfold=K,
                            metrics="logloss",objective="binary:logistic",
                            verbose=1,print_every_n=T))
which.min(boLcv$e[['test_logloss_mean']])         # 最適な反復数
## プロット
par(ps=16,mai=rep(1,4),mfrow=c(1,1))
ylim <- range(c(boLcv$e[['test_logloss_mean']],boL$e[['test_logloss']]))
plot(boLcv$e[['test_logloss_mean']],lwd=3,col=2,lty=2,type='l',ylim=ylim,log='x',xlab='反復数',ylab='対数損失')
 lines(boL$e[['test_logloss']],     lwd=3,col=1,lty=1)
legend('topleft',c('テストデータ上の対数損失','交差検証法'),col=c(1,2),lwd=c(2,2),lty=c(1,2),y.intersp=1.5,bg="white")
